module Main

where

import Data.Array ((!))
import qualified Data.ByteString.Lazy as L
import qualified Data.ByteString as S
import Data.Sequence (Seq, (<|), (|>))
import qualified Data.Sequence as Seq
import qualified Data.Foldable as F
import Text.Regex.DFA ((=~), MatchArray)
import qualified Text.Regex.DFA.ByteString as DFA
import Data.Word (Word8)
import qualified Data.Sequence.KMP as KMPSeq
import qualified Data.ByteString.KMP as K
import Data.Traversable (traverse)
import Data.List (isInfixOf)

main =
  do
    testStr <- readFile "endo.dna" >>= \s -> return $ s ++ searchStr
    lazyTestStr <- L.readFile "endo.dna" >>= \s -> return $ L.append s lazySearchStr
    strictTestStr <- S.readFile "endo.dna" >>= \s -> return $ S.append s strictSearchStr
    putStrLn $ ("Warming up files ..." ++ (show (length testStr)) ++ ", " ++ (show (L.length lazyTestStr)) ++ ", " ++ (show (S.length strictSearchStr)))
    putStrLn $ ("(strict): " ++ show ({-# SCC "main_findStrict" #-} S.findSubstring strictSearchStr strictTestStr))
    putStrLn $ ("(naive): " ++ show ({-# SCC "main_findSubstringLazy" #-} findSubstringLazy lazySearchStr lazyTestStr))
    putStrLn $ ("(kmpMatchLL): " ++ show ({-# SCC "main_kmpMatchLL" #-} K.kmpMatchLL lazySearchStr lazyTestStr))
    putStrLn $ ("(kmpMatchSS): " ++ show ({-# SCC "main_kmpMatchSS" #-} K.kmpMatchSS strictSearchStr strictTestStr))
    putStrLn $ ("(list) found: " ++ show ({-# SCC "main_findList" #-} searchStr `isInfixOf` testStr))
    putStrLn $ ("(regex) found: " ++ show ({-# SCC "main_findRegex" #-} findRegexLazy lazySearchStr lazyTestStr))
    putStrLn $ ("(lazy sequence) found: " ++ show ({-# SCC "main_findLazySequence" #-} findLazySequence lazySearchStr lazyTestStr))
    putStrLn $ ("(strict sequence) found: " ++ show ({-# SCC "main_findStrictSequence" #-} findStrictSequence strictSearchStr strictTestStr))
    putStrLn "Done!"


searchStr = "IFPIFPIFPIFPIFPIFPIFPIFP"
lazySearchStr = toLazyBS searchStr
strictSearchStr = toStrictBS searchStr

toLazyBS = L.pack . map (toEnum . fromEnum) 
toStrictBS = S.pack . map (toEnum . fromEnum)

findSubstringLazy :: L.ByteString -> L.ByteString -> Maybe Int
findSubstringLazy !test !big = go big 0
    where
        go !s !n | test `L.isPrefixOf` s = Just n
                 | L.null s              = Nothing
                 | otherwise             = go (L.tail s) (n+1)

lazyToSeq :: L.ByteString -> Seq Word8
lazyToSeq s = L.foldr (<|) Seq.empty s

strictToSeq :: S.ByteString -> Seq Word8
strictToSeq s = S.foldl (|>) Seq.empty s

findLazySequence pat str =
  let patSeq = lazyToSeq pat
      searchSeq = lazyToSeq str
  in
    KMPSeq.kmpMatch patSeq searchSeq

findStrictSequence pat str =
  let patSeq = strictToSeq pat
      searchSeq = strictToSeq str
  in
    KMPSeq.kmpMatch patSeq searchSeq

findRegexLazy :: L.ByteString -> L.ByteString -> MatchArray
findRegexLazy pat str = str =~ pat 
  
